/***********************************************************************/
/* File: bmp.c                                                         */
/*                                                                     */
/*                                                                     */
/*---------------------------------------------------------------------*/
/* Programmer: Nils Schmeisser                                         */
/*                                                                     */
/*---------------------------------------------------------------------*/
/* Revision date: 8.2.96                                               */
/*---------------------------------------------------------------------*/
/* Notes: Microsoft Windows Bitmap Speicherroutine                     */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "bmp.h"
#include "cppmacro.h"

  const int xPALETTE[40][3]={
                 { 255, 255, 255},
                 {   0,   0,   0},
                 { 255, 255,   0},
                 {  60, 251,  52},
                 { 199, 199, 199},
                 { 231, 231, 231},
                 { 109, 109, 109},
                 { 211, 211, 211},
                 { 239, 239, 239},
                 { 117, 117, 117},
                 { 178, 178, 178},
                 {  97, 162, 162},
                 { 186, 215, 215},
                 {  48,  81,  81},
                 {  81, 138, 138},
                 { 170, 170, 170},
                 { 251,  20,  65},
                 {   0,   8,   0},
                 {   0,  16,   0},
                 {   0,  24,   0},
                 {   0,  32,   0},
                 {   0,  40,   0},
                 {   0,  48,   0},
                 {   0,  56,   0},
                 {   0,  65,   0},
                 {   0,  73,   0},
                 {   0,  81,   0},
                 {   0,  89,   0},
                 {   0,  97,   0},
                 {   0, 105,   0},
                 {   0, 113,   0},
                 {   0, 121,   0},
                 {   0, 130,   0},
                 {   0, 138,   0},
                 {   0, 150,   0},
                 {   0, 158,   0},
                 {   0, 166,   0},
                 {   0, 174,   0},
                 {   0, 182,   0},
                 {   0, 190,   0}
  };

unsigned short int *get_palette(char *lutname) {
  int i;
  FILE *flut;
  unsigned short *clut;

  if (strlen(lutname)>0) {
     if ((flut=fopen(lutname,"r"))!=NULL) {
       clut=new(unsigned short,768);
       fread(&clut[  0],256*2,1,flut);
       fread(&clut[256],256*2,1,flut);
       fread(&clut[512],256*2,1,flut);
       for (i=0;i<256;i++) {
	 clut[i    ]=clut[i    ] >> 8;
	 clut[i+256]=clut[i+256] >> 8;
	 clut[i+512]=clut[i+512] >> 8;
       }
       fclose(flut);
       return clut;
     } else {
       return NULL;
     }
  } else 
    return NULL;
}

RGBRec RGB(byte r,byte g,byte b) {
  RGBRec rgb;

  rgb.RedVal  =r;
  rgb.GreenVal=g;
  rgb.BlueVal =b;

  return rgb;
}

dword swap(dword w) {
  byte b[4];

  b[0]=(0x0ff000000 & w)/0x1000000;
  b[1]=(0x000ff0000 & w)/0x10000;
  b[2]=(0x00000ff00 & w)/0x100;
  b[3]=(0x0000000ff & w);
  return b[3]*0x1000000+b[2]*0x10000+b[1]*0x100+b[0];
}

void ExportToBMP(int w,int h,byte** data,char *name,char* lutname) {
  FILE *f;
  bmp_header header;
  int x,y;
  int i,rint,gint,bint;
  unsigned short int *lut;
  byte b,b1,b2;
  byte *ba;
  byte c1,c2;
  RGBRec VGAColors[16];
     
  if (data!=NULL) {
    VGAColors[ 0]=RGB( 0x00, 0x00, 0x00 );
    VGAColors[ 1]=RGB( 0x00, 0x00, 0x80 );
    VGAColors[ 2]=RGB( 0x00, 0x80, 0x00 );
    VGAColors[ 3]=RGB( 0x00, 0x80, 0x80 );
    VGAColors[ 4]=RGB( 0x80, 0x00, 0x00 );     
    VGAColors[ 5]=RGB( 0x80, 0x00, 0x80 );
    VGAColors[ 6]=RGB( 0x80, 0x80, 0x00 );
    VGAColors[ 7]=RGB( 0x80, 0x80, 0x80 );
    VGAColors[ 8]=RGB( 0xc0, 0xc0, 0xc0 );
    VGAColors[ 9]=RGB( 0x00, 0x00, 0xff );
    VGAColors[10]=RGB( 0x00, 0xff, 0x00 );
    VGAColors[11]=RGB( 0x00, 0xff, 0xff );
    VGAColors[12]=RGB( 0xff, 0x00, 0x00 );
    VGAColors[13]=RGB( 0xff, 0x00, 0xff );
    VGAColors[14]=RGB( 0xff, 0xff, 0x00 );
    VGAColors[15]=RGB( 0xff, 0xff, 0xff );
    
    f=fopen(name,"w+");

    header.id1=0x42;
    header.id2=0x4d;
    header.tmp1=0;
    header.tmp2=swap(0x28);
    header.width=swap(w);
    header.height=swap(h);
    header.tmp4=0x0100;
    header.bpp=8; 
    for (i=0;i<3;i++) header.tmp5[i]=0;
    for (i=0;i<22;i++) header.tmp6[i]=0;
    if ((lut=get_palette(lutname))!=NULL) {
      header.pal=new(byte,256*4);
      for (i=0;i<40;i++) {
        header.pal[i*4+0]=(byte)xPALETTE[i][2];  
        header.pal[i*4+1]=(byte)xPALETTE[i][1];  
        header.pal[i*4+2]=(byte)xPALETTE[i][0]; 
        header.pal[i*4+3]=0;
      }
      for (i=40;i<256;i++) {
        rint=(int)((double)(i-40)*256.0/216.0);
        header.pal[i*4+0]=(byte)lut[rint+512]; 
        header.pal[i*4+1]=(byte)lut[rint+256]; 
        header.pal[i*4+2]=(byte)lut[rint    ]; 
        header.pal[i*4+3]=0;
      }
      delete(lut);
      header.first=1078;
    } else {
      header.pal=new(byte,16*4);
      for (i=0;i<16;i++) {
        header.pal[i*4+0]=VGAColors[i].BlueVal;
        header.pal[i*4+1]=VGAColors[i].GreenVal;
        header.pal[i*4+2]=VGAColors[i].RedVal;
        header.pal[i*4+3]=0;
      }
      header.first=118;
    } 
    header.flenl=swap((dword)header.first+(dword)w*(dword)h);
    header.first=swap(header.first);

    fwrite(&(header.id1),1,1,f);
    fwrite(&(header.id2),1,1,f);
    fwrite(&(header.flenl),4,1,f);
    fwrite(&(header.tmp1),4,1,f);
    fwrite(&(header.first),4,1,f);
    fwrite(&(header.tmp2),4,1,f);
    fwrite(&(header.width),4,1,f);
    fwrite(&(header.height),4,1,f);
    fwrite(&(header.tmp4),2,1,f);
    fwrite(&(header.bpp),1,1,f);
    fwrite(header.tmp5,3,1,f);
    fwrite(header.tmp6,22,1,f);

    header.first=swap(header.first);
    for (i=0;i<header.first-54;i++) fwrite(&(header.pal[i]),1,1,f);
    delete(header.pal);
     
    for (y=h-1;y>=0;y--) {
      fwrite(data[y],w,1,f);
    }
    fclose(f);
  }
}

void ExportToAVSFLD(int w,int h,byte** data,char *name) {
  FILE *f;
  int i;
  
  f=fopen(name,"w+");
  fprintf(f,"# AVS field file\n");
  fprintf(f,"ndim = 2\n");
  fprintf(f,"dim1 = %d\n",h);
  fprintf(f,"dim2 = %d\n",w);
  fprintf(f,"nspace = 2\n");
  fprintf(f,"veclen = 1\n");
  fprintf(f,"data = byte\n");
  fprintf(f,"field = uniform\n");
  fprintf(f,"%c%c",0x0c,0x0c);

  for (i=0;i<h;i++) fwrite(data[i],1,w,f);
  fclose(f);
}
