#include <spp_prog_model.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "cppmacro.h"
#include "bmp.h"

void mandel(const double g,const double X1,const double Y2,
	    const double xd,const double yd,
	    const int maxx,const int maxy,const int maxc,const int z,
	    byte **buffer) {

  // Kapseln der, bzgl. eines Threads lokalen Variablen
  // Dies ist wichtig, um Fremdzugriffe zu verhindern.

  static thread_private int i,j,k,n,h;
  static thread_private double a,b,m,x,y;

  // Parallelisierung der auesseren Schleife erzwingen
  // und nur soviel Threads (max_threads=4) wie physisch vorhandene CPU's zulassen
  // Zu beachten ist Privatisierung der Variablen der inneren Schleifen => Datenkapselung

  #pragma _CNX loop_parallel(ivar=i,max_threads=4)
  #pragma _CNX loop_private(j,y,x,a,b,m,h)
  for (i=0;i<maxy;i++) {
    y=Y2-i*yd;

    // Explizit Sequentialitaet anweisen

    #pragma _CNX no_parallel
    for (j=0;j<maxx;j++) {
      x=X1+j*xd;
      a=0;
      b=0;
      h=0;
      #pragma _CNX no_parallel
      for (n=1;n<=z;n++) {
	m=a*a-b*b+x;
	b=2*a*b+y;
	a=m;
	if (fabs(a)+fabs(b)>g) {
	  h=n/2;
	  n=z;
	}
      }

      // Da buffer nicht als loop_private deklariert werden kann (=> segmentation fault)
      // muesste die nachfolgende Anweisung eigentlich als kritischer Bereich deklariert
      // werden (konkurrierende Schreibzugriffe der Threads). Da aber klar ist, das keine
      // zwei Threads gleichzeitig auf ein und das selbe Feldelement zugreifen, kann man
      // dies vernachlaessigen und die Meldung des Compilers ignorieren.

      buffer[i][j]=(byte)(h % maxc);
    }

  }
}

int main(int argc,char **argv) {
  int drv,mode;
  double x0,y0,size;
  int z,maxx,maxy,maxc;
  double g,X1,Y2;
  double xd,yd;
  int i;
  static node_private byte **buffer;

  printf("ex3: Schleifenparallelisierung, Threadkonzept und Datenkapselung.\n");
  printf("     Das Programm berechnet einen Ausschnitt aus der Mandelbrot-Menge und\n");
  printf("     speichert das Ergebnis in der Datei 'mandel.bmp'.\n");
  printf("     (c) N. Schmeisser, FVTK, FZ Rossendorf, 1996\n");

  printf(" Mittelpunkt und Groesse: ");
  scanf("%lf %lf %lf",&x0,&y0,&size);
  printf("                  Grenze:");
  scanf("%fl",&g);
  printf("Maximale Iterationstiefe:");
  scanf("%d",&z);

  maxx=480;
  maxy=480;
  maxc=256;

  X1=x0-size/2.0;
  Y2=y0+size/2.0;
  xd=size/(double)maxx;	
  yd=size/(double)maxy;


  buffer=sppnew(byte*,maxy,NODE_PRIVATE_MEM);
  for (i=0;i<maxy;i++)
    buffer[i]=sppnew(byte,maxx,NODE_PRIVATE_MEM);

  mandel(g,X1,Y2,xd,yd,maxx,maxy,maxc,z,buffer);

  ExportToBMP(maxx,maxy,buffer,"mandel.bmp","red.lut");

  for (i=0;i<maxy;i++)
    delete(buffer[i]);
  delete(buffer);
  
  printf("Ready.\n");
  return 0;
}





