#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <spp_prog_model.h>
#include <cps.h>

#define new(TYPEID,SIZE) ( (TYPEID*)malloc(SIZE*sizeof(TYPEID)) )
#define delete(PTRINST) (free(PTRINST))

typedef struct {
          int from;
          int to;
	  unsigned long msgl;
	  char *msg;
	  int lock;
	} message_t;

static node_private gate_t gate;
static node_private message_t msg;
static node_private ntasks;

void philosoph(void *arg) {

  // i_am muss gekapselt werden, da i_am sonst ueberschrieben werden koennte

  static thread_private int i_am;
  
  // Ermitteln der Thread ID

  i_am=cps_stid();
  printf("Philosoph %d ist da.\n",i_am);

  // Endlosschleife

  while (1) {

    // Anfragen, ob kritischer Bereich betreten werden darf, wenn nicht, wird
    // der Thread an dieser Stelle angehalten und wartet, bis das Semaphor den
    // Zugang oeffnet

    lock_gate(&gate);

    // Zugang ist jetzt fuer andere Threads gesperrt

      // pruefen ob message Puffer leer ist

      if (!(msg.lock)) {

        // message Puffer besetzten und Absetzen einer Nachricht vorbereiten

	msg.lock=1;
	msg.from=i_am;
	msg.to=rand() % ntasks;
	sprintf(msg.msg,"Philosoph %2d gruesst Philosoph %2d.",i_am,msg.to);
      } else if (msg.to==i_am) {

        // message puffer war voll und der Inhalt fuer diesen Thread bestimmt

	printf("Philosoph %2d hoert von Philosoph %2d: '%s'\n",i_am,msg.from,msg.msg);

        // message Puffer freigeben

	msg.lock=0;
      }

    // kritischen Bereich verlassen

    unlock_gate(&gate);

    // kritischer Bereich wurde verlassen und freigegeben

  }
}

int main(int argc,char **argv) {
  spawn_sym_t tinfo;

  printf("ex6: Dynamische Erzeugung von Threads (Anzahl erst zur Laufzeit bekannt,\n");
  printf("     symmetrischer Aufruf, kritische Bereiche, Kommunikation.\n");
  printf("     Falls das Programm haengt, bitte mit:\n");
  printf("     'mpa -m -min <neuanzahl> -max <neuanzahl> ex6' die Anzahl der Threads\n");
  printf("     erhoehen.\n");
  printf("     Das Programm soll irgendwann mal das Philosophenproblem zeigen.\n");
  printf("     Die Philosophen sind schon da, reden aber nur miteinander.\n");
  printf("     Das Programm muss mit CTRL-C abgebrochen werden.\n");
  printf("     (c) N. Schmeisser, FVTK, FZ Rossendorf, 1996\n");

  // gate ist eine Art -> Semaphor (Signal). Es zeiht anderen Threads an, dass eine Bereich
  // schon betreten wurde

  alloc_gate(&gate);
  printf("Anzahl Philosophen:");
  scanf("%d",&ntasks);

  // Nachrichtenpuffer vorbereiten

  msg.msg=new(char,256);

  // die symmetrischen Erzeugung von ntasks Threads vorbereiten

  tinfo.node=-2;
  tinfo.min=ntasks;
  tinfo.max=ntasks;
 
  // CPS_OVER_SUBSCRIBE sagt dem Compiler, dass mehrere Threads auf einer CPU laufen
  // koennen. Der Compiler muss auf jeden Fall die Option -Wl,+over an den Linker
  // weiterleiten.

  tinfo.threadscope=CPS_OVER_SUBSCRIBE || CPS_THREAD_PARALLEL;

  // Aufruf und Ausgabe des Fehlerstatus  

  printf("Status: %d, %d\n",cps_ppcall(&tinfo,philosoph,NULL),errno);

  delete(msg.msg);
  return 0;
}
